import json
from copy import deepcopy
from datetime import datetime, timezone
from typing import Type, Optional, NamedTuple, Dict, Any, Tuple, List
from unittest.mock import patch

import pandas as pd
import numpy as np

import pytest
import responses
from dateutil.parser import isoparse

from arthurai.core.models import ArthurModel
from arthurai import util as arthur_util

from tests.helpers import mock_post, MockDatetime, MockShortUUID, mock_patch, assert_kwargs_equal
from tests.fixtures.mocks import client
from tests.fixtures.models.regression import batch_model, INT_INPUT, FLOAT_NONINPUT, PRED, GROUND_TRUTH, TABULAR_MODEL_ID
from tests.fixtures.models.multiclass import biclass_model

# FIXTURES
INT_INPUT_VAL_1 = 1
INT_INPUT_VAL_2 = 2
FLOAT_INPUT_VAL_1 = 3.0
FLOAT_INPUT_VAL_2 = 4.0
PRED_VAL_1 = 5.0
PRED_VAL_2 = 10.0
GROUND_TRUTH_VAL_1 = 5.5
GROUND_TRUTH_VAL_2 = 9.5
INFERENCE_ID_1 = "inf1"
INFERENCE_ID_2 = "inf2"
TIMESTAMP_VALUE = '2021-01-21T12:00:00.123456Z'
BATCH_ID = 'batch_1'
BATCH_ID_2 = 'batch_2'

API_FORMAT_DATA = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_1,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
            PRED: PRED_VAL_1
        },
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_1
        },
        'batch_id': BATCH_ID
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_2,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
            PRED: PRED_VAL_2
        },
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_2
        },
        'batch_id': BATCH_ID
    }
]

API_FORMAT_DATA_NO_BATCH = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_1,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
            PRED: PRED_VAL_1
        },
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_1
        }
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_2,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
            PRED: PRED_VAL_2
        },
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_2
        }
    }
]

API_FORMAT_DATA_NO_PREDS_OR_GROUND_TRUTH = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_1,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
        },
        'batch_id': BATCH_ID
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_2,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
        },
        'batch_id': BATCH_ID
    }
]

API_FORMAT_DATA_NO_GROUND_TRUTH = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_1,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
            PRED: PRED_VAL_1
        },
        'batch_id': BATCH_ID
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_2,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
            PRED: PRED_VAL_2
        },
        'batch_id': BATCH_ID
    }
]

API_FORMAT_DATA_INFERENCES_ONLY = [
    {
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_1,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
        }
    },
    {
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_2,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
        }
    }
]

API_FORMAT_DATA_PARTIAL_GT = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_1,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
            PRED: PRED_VAL_1
        },
        'batch_id': BATCH_ID
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_2,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
            PRED: PRED_VAL_2
        },
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_2
        },
        'batch_id': BATCH_ID
    }
]

FLAT_DATA_FULL_INPUTS = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'inference_timestamp': TIMESTAMP_VALUE,
        INT_INPUT: INT_INPUT_VAL_1,
        FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
        PRED: PRED_VAL_1,
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        GROUND_TRUTH: GROUND_TRUTH_VAL_1,
        'batch_id': BATCH_ID
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'inference_timestamp': TIMESTAMP_VALUE,
        INT_INPUT: INT_INPUT_VAL_2,
        FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
        PRED: PRED_VAL_2,
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        GROUND_TRUTH: GROUND_TRUTH_VAL_2,
        'batch_id': BATCH_ID
    }
]

FLAT_DATA_NO_TIMESTAMPS_OR_IDS = [
    {
        INT_INPUT: INT_INPUT_VAL_1,
        FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
        PRED: PRED_VAL_1,
        GROUND_TRUTH: GROUND_TRUTH_VAL_1,
        'batch_id': BATCH_ID
    },
    {
        INT_INPUT: INT_INPUT_VAL_2,
        FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
        PRED: PRED_VAL_2,
        GROUND_TRUTH: GROUND_TRUTH_VAL_2,
        'batch_id': BATCH_ID
    }
]

FLAT_DATA_INFERENCES_ONLY = [
    {
        INT_INPUT: INT_INPUT_VAL_1,
        FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
        PRED: PRED_VAL_1,
    },
    {
        INT_INPUT: INT_INPUT_VAL_2,
        FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
        PRED: PRED_VAL_2,
    }
]

FLAT_DATA_PARTIAL_GT = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'inference_timestamp': TIMESTAMP_VALUE,
        INT_INPUT: INT_INPUT_VAL_1,
        FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
        PRED: PRED_VAL_1,
        'batch_id': BATCH_ID
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'inference_timestamp': TIMESTAMP_VALUE,
        INT_INPUT: INT_INPUT_VAL_2,
        FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
        PRED: PRED_VAL_2,
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        GROUND_TRUTH: GROUND_TRUTH_VAL_2,
        'batch_id': BATCH_ID
    }
]

API_FORMAT_GROUND_TRUTHS = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_1
        }
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_2
        }
    }
]

FLAT_GROUND_TRUTH_DATA_FULL_INPUTS = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        GROUND_TRUTH: GROUND_TRUTH_VAL_1
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        GROUND_TRUTH: GROUND_TRUTH_VAL_2
    }
]

FLAT_GROUND_TRUTH_DATA_NO_TIMESTAMPS = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        GROUND_TRUTH: GROUND_TRUTH_VAL_1
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        GROUND_TRUTH: GROUND_TRUTH_VAL_2
    }
]

MULTI_BATCH_DATA = [
    {
        'partner_inference_id': INFERENCE_ID_1,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_1,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_1,
            PRED: PRED_VAL_1
        },
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_1
        },
        'batch_id': BATCH_ID
    },
    {
        'partner_inference_id': INFERENCE_ID_2,
        'inference_timestamp': TIMESTAMP_VALUE,
        'inference_data': {
            INT_INPUT: INT_INPUT_VAL_2,
            FLOAT_NONINPUT: FLOAT_INPUT_VAL_2,
            PRED: PRED_VAL_2
        },
        'ground_truth_timestamp': TIMESTAMP_VALUE,
        'ground_truth_data': {
            GROUND_TRUTH: GROUND_TRUTH_VAL_2
        },
        'batch_id': BATCH_ID_2
    }
]

BATCH_COUNTS = {
    BATCH_ID: 2
}

MULTI_BATCH_COUNTS = {
    BATCH_ID: 1,
    BATCH_ID_2: 1
}

PREDS_RAW_LIST = [PRED_VAL_1, PRED_VAL_2]
GROUND_TRUTH_RAW_LIST = [GROUND_TRUTH_VAL_1, GROUND_TRUTH_VAL_2]

FULL_DATA_FRAME = pd.DataFrame(FLAT_DATA_FULL_INPUTS)
FULL_DATA_FRAME['inference_timestamp'] = pd.to_datetime(FULL_DATA_FRAME['inference_timestamp'])
FULL_DATA_FRAME['ground_truth_timestamp'] = pd.to_datetime(FULL_DATA_FRAME['ground_truth_timestamp'])

EXPECTED_RESPONSE = {
    "counts": {
        "success": 2,
        "failure": 0,
        "total": 2
    },
    "results": [
        {
            "partner_inference_id": INFERENCE_ID_1,
            "message": "success",
            "status": 200
        },
        {
            "partner_inference_id": INFERENCE_ID_2,
            "message": "success",
            "status": 200
        }
    ]
}

EXPECTED_RESPONSE_NO_IDS = {
    "counts": {
        "success": 2,
        "failure": 0,
        "total": 2
    },
    "results": [
        {
            "message": "success",
            "status": 200
        },
        {
            "message": "success",
            "status": 200
        }
    ]
}


@pytest.fixture(params=['batch_model', 'biclass_model'])
def model(request):
    return request.getfixturevalue(request.param)


class TestInferences:

    def test_format_inference_no_gt(self, client, batch_model):
        inference = batch_model._format_inference_request(
            inference_timestamp=datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
            partner_inference_id="1234",
            model_pipeline_input=ArthurModel._replace_nans_and_infinities_in_dict({"attr": 5}),
            non_input_data=ArthurModel._replace_nans_and_infinities_in_dict({"Gender": "male"}),
            predicted_value=ArthurModel._replace_nans_and_infinities_in_dict({"prediction": 800}),
            ground_truth=ArthurModel._replace_nans_and_infinities_in_dict(None)
        )

        inferences = arthur_util.format_timestamps([inference])
        assert len(inferences) == 1
        assert inferences[0]["inference_timestamp"] == "2020-08-13T17:44:31Z"
        assert "ground_truth_timestamp" not in inferences[0]
        assert "ground_truth" not in inferences[0]

    @responses.activate
    def test_format_inference_request(self, batch_model):
        inf = batch_model._format_inference_request(
            inference_timestamp="2020-08-13T17:44:31.552125Z",
            partner_inference_id="1234",
            model_pipeline_input={"attr": 5},
            non_input_data={"Gender": "male"},
            predicted_value={"prediction": 800},
            ground_truth={"actual_value": 812}
        )

        expected = {
            "inference_timestamp": "2020-08-13T17:44:31.552125Z",
            "partner_inference_id": "1234",
            "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
            "ground_truth_timestamp": "2020-08-13T17:44:31.552125Z",
            "ground_truth_data": {"actual_value": 812}
        }

        assert inf == expected

    def test_update_inference_timestamp(self):
        inference_incorrect_timestamps = [
            {
                "inference_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "partner_inference_id": "1234",
                "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
                "ground_truth_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "ground_truth_data": {"actual_value": 812}
            },
            {
                "inference_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "partner_inference_id": "1234",
                "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
                "ground_truth_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "ground_truth_data": {"actual_value": 812}
            }
        ]

        reformatted_inferences = arthur_util.format_timestamps(inference_incorrect_timestamps)
        for inf in reformatted_inferences:
            assert inf["inference_timestamp"] == inf["ground_truth_timestamp"] == "2020-08-13T17:44:31Z"

    def test_string_update_inference_timestamp(self):
        inference_incorrect_timestamps = [
            {
                "inference_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "partner_inference_id": "1234",
                "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
                "ground_truth_timestamp": "2020-08-13T17:44:31.552125",
                "ground_truth_data": {"actual_value": 812}
            },
            {
                "inference_timestamp": "2020-08-13T17:44:31.552125",
                "partner_inference_id": "1234",
                "inference_data": {"attr": 5, "Gender": "male", "prediction": 800},
                "ground_truth_timestamp": datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
                "ground_truth_data": {"actual_value": 812}
            }
        ]

        reformatted_inferences = arthur_util.format_timestamps(inference_incorrect_timestamps)
        for inf in reformatted_inferences:
            assert inf["inference_timestamp"] in ["2020-08-13T17:44:31Z", "2020-08-13T17:44:31.552125Z"]
            assert inf["ground_truth_timestamp"] in ["2020-08-13T17:44:31Z", "2020-08-13T17:44:31.552125Z"]

        # TODO: uncomment this block and delete one above to depreciate support for strings
        # for obj in inference_incorrect_timestamps:
        #     with pytest.raises(TypeError):
        #         arthur_util.format_timestamps([obj])

    def test_format_timestamp_with_location(self):
        unaware_timestamp = datetime(2021, 2, 8, 12, 10, 5)
        assert arthur_util.format_timestamp(unaware_timestamp, "America/Phoenix") == "2021-02-08T19:10:05Z"

    # this is actually more generic but leaving scoped this way for now
    class InferencesCase(NamedTuple):
        pos_args: Tuple
        kwargs: Dict[str, Any]
        expected_request: List[Dict[str, Any]]
        expected_response: Dict[str, Any]
        batch_counts: Dict[str, int] = {}

    send_inferences_cases = [
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': API_FORMAT_DATA}, expected_request=API_FORMAT_DATA,
            expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS), id="orig-api-format"),
        pytest.param(InferencesCase(
            pos_args=(API_FORMAT_DATA,), kwargs={}, expected_request=API_FORMAT_DATA,
            expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS), id="orig-api-format-positional"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': API_FORMAT_DATA_NO_PREDS_OR_GROUND_TRUTH, 'predictions': PREDS_RAW_LIST,
                                 'ground_truths': GROUND_TRUTH_RAW_LIST},
            expected_request=API_FORMAT_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="api-format-with-preds-and-gt-separate-no-gt-timestamps"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': API_FORMAT_DATA_NO_PREDS_OR_GROUND_TRUTH,
                                 'predictions': PREDS_RAW_LIST},
            expected_request=API_FORMAT_DATA_NO_GROUND_TRUTH, expected_response=EXPECTED_RESPONSE,
            batch_counts=BATCH_COUNTS), id="api-format-no-ground-truth"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': FLAT_DATA_FULL_INPUTS}, expected_request=API_FORMAT_DATA,
            expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS), id="flat-dict-format"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': FLAT_DATA_INFERENCES_ONLY,
                                 'predictions': PREDS_RAW_LIST,
                                 'ground_truths': GROUND_TRUTH_RAW_LIST,
                                 'batch_id': BATCH_ID},
            expected_request=API_FORMAT_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="flat-dict-format-separate-preds-gt"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': FLAT_DATA_INFERENCES_ONLY,
                                 'predictions': PREDS_RAW_LIST,
                                 'ground_truths': GROUND_TRUTH_RAW_LIST},
            expected_request=API_FORMAT_DATA_NO_BATCH, expected_response=EXPECTED_RESPONSE, batch_counts={}),
            id="flat-dict-format-separate-preds-gt-no-batch"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': FLAT_DATA_PARTIAL_GT,
                                 'batch_id': BATCH_ID},
            expected_request=API_FORMAT_DATA_PARTIAL_GT, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="flat-dict-format-partial-gt"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': pd.DataFrame(FLAT_DATA_INFERENCES_ONLY),
                                 'predictions': pd.Series(PREDS_RAW_LIST),
                                 'ground_truths': np.array(GROUND_TRUTH_RAW_LIST),
                                 'batch_id': BATCH_ID},
            expected_request=API_FORMAT_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="mixed-types"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': FULL_DATA_FRAME},
            expected_request=API_FORMAT_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="single-data-frame"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': pd.DataFrame(FLAT_DATA_INFERENCES_ONLY),
                                 'predictions': pd.DataFrame({PRED: PREDS_RAW_LIST}),
                                 'ground_truths': pd.DataFrame({GROUND_TRUTH: GROUND_TRUTH_RAW_LIST}),
                                 'batch_id': BATCH_ID},
            expected_request=API_FORMAT_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="three-dataframes"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': FLAT_DATA_NO_TIMESTAMPS_OR_IDS,
                                 'batch_id': BATCH_ID,
                                 'inference_timestamps': [TIMESTAMP_VALUE, TIMESTAMP_VALUE],
                                 'ground_truth_timestamps': [TIMESTAMP_VALUE, TIMESTAMP_VALUE],
                                 'partner_inference_ids': [INFERENCE_ID_1, INFERENCE_ID_2]},
            expected_request=API_FORMAT_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="timestamps-and-ids-separate"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': FLAT_DATA_NO_TIMESTAMPS_OR_IDS,
                                 'batch_id': BATCH_ID,
                                 'inference_timestamps': pd.Series([TIMESTAMP_VALUE, TIMESTAMP_VALUE]),
                                 'ground_truth_timestamps': pd.Series([TIMESTAMP_VALUE, TIMESTAMP_VALUE]),
                                 'partner_inference_ids': pd.Series([INFERENCE_ID_1, INFERENCE_ID_2])},
            expected_request=API_FORMAT_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="timestamps-and-ids-separate-series"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': FLAT_DATA_NO_TIMESTAMPS_OR_IDS,
                                 'batch_id': BATCH_ID,
                                 'inference_timestamps': [isoparse(TIMESTAMP_VALUE),
                                                          isoparse(TIMESTAMP_VALUE)],
                                 'ground_truth_timestamps': [isoparse(TIMESTAMP_VALUE), isoparse(TIMESTAMP_VALUE)],
                                 'partner_inference_ids': [INFERENCE_ID_1, INFERENCE_ID_2]},
            expected_request=API_FORMAT_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=BATCH_COUNTS),
            id="timestamps-and-ids-separate-datetime-format"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'inferences': MULTI_BATCH_DATA},
            expected_request=MULTI_BATCH_DATA, expected_response=EXPECTED_RESPONSE, batch_counts=MULTI_BATCH_COUNTS),
            id="multi-batch"),

    ]

    @responses.activate
    @pytest.mark.parametrize("case", send_inferences_cases)
    def test_send_inferences(self, model, case: InferencesCase):
        pos_args_copy = deepcopy(case.pos_args)
        kwargs_copy = deepcopy(case.kwargs)
        model_id = model.id
        mock_post(f"/api/v3/models/{model_id}/inferences", case.expected_response, status=207)
        for batch_id in case.batch_counts.keys():
            mock_patch(f"/api/v3/models/{model_id}/batches/{batch_id}", {'status': "ok"}, status=200)
        uuid_generator = MockShortUUID([INFERENCE_ID_1, INFERENCE_ID_2])
        datetime_generator = MockDatetime([TIMESTAMP_VALUE])

        with patch("arthurai.core.models.shortuuid.uuid") as uuid_mock:
            with patch("arthurai.core.models.datetime") as datetime_mock:
                uuid_mock.side_effect = uuid_generator.next
                datetime_mock.now.side_effect = datetime_generator.next
                actual_response = model.send_inferences(*case.pos_args, **case.kwargs)

            # inputs were not modified
            assert case.pos_args == pos_args_copy
            assert_kwargs_equal(kwargs_copy, case.kwargs)

            assert len(responses.calls) == 1 + len(case.batch_counts)
            assert responses.calls[0].request.body is not None
            assert responses.calls[0].request.body is not None

            actual_request_body = json.loads(responses.calls[0].request.body)
            assert actual_request_body == case.expected_request
            assert actual_response == case.expected_response

            # batch closed
            batch_close_calls = {}
            for call in list(responses.calls)[1:]:
                assert call.request.url is not None
                batch_close_calls[call.request.url.split("/")[-1]] = call.request.body
            for batch_id in case.batch_counts.keys():
                expected_body = json.dumps({
                    "status": "uploaded",
                    "total_record_count": case.batch_counts[batch_id]
                })
                assert batch_close_calls[batch_id] == expected_body

    update_ground_truths_cases = [
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'ground_truths': API_FORMAT_GROUND_TRUTHS}, expected_request=API_FORMAT_GROUND_TRUTHS,
            expected_response=EXPECTED_RESPONSE_NO_IDS), id="orig-api-format"),
        pytest.param(InferencesCase(
            pos_args=(API_FORMAT_GROUND_TRUTHS,), kwargs={}, expected_request=API_FORMAT_GROUND_TRUTHS,
            expected_response=EXPECTED_RESPONSE_NO_IDS), id="orig-api-format-positional"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'ground_truths': FLAT_GROUND_TRUTH_DATA_FULL_INPUTS},
            expected_request=API_FORMAT_GROUND_TRUTHS, expected_response=EXPECTED_RESPONSE_NO_IDS),
            id="flat-dict-format"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'ground_truths': FLAT_GROUND_TRUTH_DATA_NO_TIMESTAMPS},
            expected_request=API_FORMAT_GROUND_TRUTHS, expected_response=EXPECTED_RESPONSE_NO_IDS),
            id="flat-dict-no-timestamps"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'ground_truths': GROUND_TRUTH_RAW_LIST,
                                 'partner_inference_ids': [INFERENCE_ID_1, INFERENCE_ID_2]},
            expected_request=API_FORMAT_GROUND_TRUTHS, expected_response=EXPECTED_RESPONSE_NO_IDS),
            id="single-list"),
        pytest.param(InferencesCase(
            pos_args=(), kwargs={'ground_truths': pd.DataFrame(FLAT_GROUND_TRUTH_DATA_FULL_INPUTS)},
            expected_request=API_FORMAT_GROUND_TRUTHS, expected_response=EXPECTED_RESPONSE_NO_IDS),
            id="single-dataframe")
    ]

    @responses.activate
    @pytest.mark.parametrize("case", update_ground_truths_cases)
    def test_update_ground_truth(self, model, case: InferencesCase):
        pos_args_copy = deepcopy(case.pos_args)
        kwargs_copy = deepcopy(case.kwargs)
        mock_patch(f"/api/v3/models/{model.id}/inferences", case.expected_response, status=207)
        datetime_generator = MockDatetime([TIMESTAMP_VALUE])

        with patch("arthurai.core.models.datetime") as datetime_mock:
            datetime_mock.now.side_effect = datetime_generator.next
            actual_response = model.update_inference_ground_truths(*case.pos_args, **case.kwargs)

        # inputs were not modified
        assert case.pos_args == pos_args_copy
        assert_kwargs_equal(kwargs_copy, case.kwargs)

        assert len(responses.calls) == 1
        assert responses.calls[0].request.body is not None
        actual_request_body = json.loads(responses.calls[0].request.body)

        assert actual_request_body == case.expected_request
        assert actual_response == case.expected_response



from functools import partial
from typing import Any, Dict, List, Tuple

import gorilla
from docutils.parsers.rst import Directive
from docutils.statemachine import ViewList
from sphinx.application import Sphinx

from sphinx.ext.autosummary import generate
from sphinx.ext import autosummary
from docutils import nodes
from sphinx.util import nested_parse_with_titles
from tabulate import tabulate

"""Custom Permissions Table generation extension.

WARNING: this implementation is heinous. It stores state on the class. It monkey-patches third-party library code.
It should be a crime, but it works as intended.

This module is used in place of "sphinx.ext.autosummary" (declared in conf.py). It calls out to sphinx.ext.autosummary 
to provide all of the regular old autosummary functionality, but also patches a couple bits of its functionality to 
view the scanned objects and store any :permissions: fields their docstrings contain.
"""


class PermissionGenerator(Directive):

    """Sphinx directive to create permissions tables.

    Provides functionality to patch autosummary's template_render() and generate_autosummary_content() functions so
    that we can grab any :permissions: fields on the scanned methods and functions, and store their contents to later
    fill in the table.

    Unfortunately as a Directive this object doesn't get instantiated until well after autosummary has completed its
    scan. This means that we need to store the contents of the scanned objects on the class itself <vomit.jpg>.
    """

    permissions: Dict[str, Tuple[str, str]] = {}
    patched_templates: List[generate.AutosummaryRenderer] = []
    cur_obj: Any = None

    @classmethod
    def template_render(cls, template: generate.AutosummaryRenderer, template_name: str, context: Dict) -> str:
        """
        Patches autosummary's AutosummaryRenderer.render() method, capturing any :permissions: fields of the
        relevant objects.
        """
        if context['objtype'] == "function":
            cls.record_permissions(cls.cur_obj, context['fullname'], "function")
        elif context['objtype'] == "class":
            for method_name in context['methods']:
                full_method_name = context['fullname'] + "." + method_name
                cls.record_permissions(getattr(cls.cur_obj, method_name), full_method_name, "method")
        actual_template_render = gorilla.get_original_attribute(template, 'render')
        return actual_template_render(template_name, context)

    @classmethod
    def generate_autosummary_content(cls, name: str, obj: Any, parent: Any, template: generate.AutosummaryRenderer,
                                     template_name: str, imported_members: bool, app: Any,
                                     recursive: bool, context: Dict, modname: str = None, qualname: str = None) -> str:
        """
        Patches autosummary's generate_autosummary_content() function, storing the current object to be rendered by
        the AutosummaryRenderer.render() method. This patches that method so that we can store the permissions of the
        needed functions and methods.
        """
        # patch the incoming template to use our render function
        # this lets us dynamically grab the values being passed into the template, such as an object's functions,
        # attributes, etc
        if template not in cls.patched_templates:
            patch = gorilla.Patch(template, 'render', partial(cls.template_render, template),
                                  settings=gorilla.Settings(allow_hit=True))
            gorilla.apply(patch)
            cls.patched_templates.append(template)

        # we'll want to access those functions and methods by name on the object, so we need to store the object
        cls.cur_obj = obj

        # now we'll make the call to the original generate_autosummary_content(). after populating the dict
        # of functions, methods, attributes, etc. on the object it will call template.render() with them, which will
        # redirect to our patched render() where we can inspect them for permission fields
        actual_generate_autosummary_content = gorilla.get_original_attribute(generate, 'generate_autosummary_content')
        return actual_generate_autosummary_content(
            name, obj, parent, template, template_name, imported_members, app, recursive, context, modname, qualname)

    @classmethod
    def record_permissions(cls, obj: Any, name: str, objtype: str) -> None:
        """
        Given an object, if it contains a :permissions: field add it to a running dictionary
        :param obj: the object to check
        :param name: the full name of the object
        :param objtype: "function" or "method"
        :return: None, side-effects the PermissionGenerator object
        """
        if obj.__doc__ is None:
            return
        doc_lines = obj.__doc__.split("\n")
        # just find all lines that start with ":permissions:"
        permissions_lines = [(i, line.strip()[len(":permissions:"):].strip()) for i, line in enumerate(doc_lines)
                             if line.strip().startswith(":permissions:")]
        if len(permissions_lines) == 0:
            return
        elif len(permissions_lines) == 1:
            # ensure that the permissions field is only a single line
            line_num, perm_line = permissions_lines[0]
            next_line_stripped = doc_lines[line_num + 1].strip()
            if next_line_stripped != "" and not next_line_stripped.startswith(":"):
                raise ValueError(f"Sorry, multiline :permissions: fields aren't supported! The line following the "
                                 f":permissions: field in {name} must be blank or a new field (e.g. :raises:)")
            # add this permission to our running list
            cls.permissions[name] = (perm_line, objtype)
        else:
            raise ValueError(f"{name} has {len(permissions_lines)} :permissions: fields. There should be only a single "
                             f"permissions field per object.")

    @classmethod
    def create_permission_table(cls) -> str:
        # generate permissions table as RST
        table = []
        for name, (perm, objtype) in cls.permissions.items():
            if objtype == "function":
                link = f":py:func:`{name}`"
            elif objtype == "method":
                link = f":py:meth:`{name}`"
            else:
                raise ValueError(f"unexpected object type {objtype}")
            table.append([link, perm])

        return tabulate(table, headers=["Function/Method", "Permissions"], tablefmt="rst")

    def run(self) -> List[nodes.Node]:
        # get the RST table
        table_content = self.create_permission_table()

        # parse the RST into docutils nodes to add the table to the document
        # from https://stackoverflow.com/a/44084890
        rst = ViewList()
        for i, line in enumerate(table_content.split("\n")):
            rst.append(line, self.state.document.current_source, i)

        node = nodes.section()
        node.document = self.state.document

        nested_parse_with_titles(self.state, rst, node)

        return node.children


def setup(app: Sphinx):
    # patch autosummary's generate_autosummary_content function
    # this will ensure that we get to capture inputs and outputs and wrap the logic any time this function is
    # called to generate a stub for a module/class/function, at which time we'll scan for :permissions: fields
    gorilla.apply(gorilla.Patch(generate, 'generate_autosummary_content',
                                PermissionGenerator.generate_autosummary_content,
                                settings=gorilla.Settings(allow_hit=True)))

    app.add_directive("permissions-table", PermissionGenerator)

    # call out to autosummary setup to properly register all the classic autosummary functionality
    autosummary.setup(app)



import pytest
import mock
import os

from arthurai.client.client import new_requests_client
from arthurai.core.models import ArthurModel, ArthurAttribute
from arthurai.client.http.requests import HTTPClient
from tests.base_test import BaseTest


@pytest.fixture(scope="function")
def mock_cred_env_vars():
    with mock.patch.dict(os.environ, {"ARTHUR_API_KEY": "access_key", "ARTHUR_ENDPOINT_URL": BaseTest._base_url, "VERIFY_SSL": "False"}):
        yield


@pytest.fixture(scope="function")
def mock_cred_env_vars_ssl_true():
    with mock.patch.dict(os.environ, {"ARTHUR_API_KEY": "access_key", "ARTHUR_ENDPOINT_URL": BaseTest._base_url, "VERIFY_SSL": "True"}):
        yield


@pytest.fixture(scope="function")
def mock_cred_env_vars_ssl_garbage():
    with mock.patch.dict(os.environ, {"ARTHUR_API_KEY": "access_key", "ARTHUR_ENDPOINT_URL": BaseTest._base_url, "VERIFY_SSL": "bernie"}):
        yield


@pytest.fixture(scope="function")
def binary_classification_model() -> ArthurModel:
    binary_classification_model = {
        "id": "4a57c553-e787-4307-88f1-3747ec9130f5",
        "partner_model_id": "binary_classification_model",
        "input_type": "TABULAR",
        "output_type": "MULTICLASS",
        "archived": False,
        "created_at": "2020-10-26T21:43:24.632335Z",
        "updated_at": "2020-10-26T21:43:26.364596Z",
        "attributes": [
            ArthurAttribute(
                id="1d4bec8d-84a6-4ae7-a7ea-1a2230bb977c",
                name="input",
                value_type="INTEGER",
                stage="PIPELINE_INPUT",
                position=1,
                categorical=False,
                monitor_for_bias=False,
                is_positive_predicted_attribute=False
            ),
            ArthurAttribute(
                id="254c8f5f-f4b8-4943-89bf-db106d5ca835",
                name="pos_predicted_probability",
                value_type="FLOAT",
                stage="PREDICTED_VALUE",
                position=1,
                categorical=False,
                min_range=0,
                max_range=1,
                monitor_for_bias=False,
                is_unique=False,
                is_positive_predicted_attribute=True,
                attribute_link="pos_gt"
            ),
            ArthurAttribute(
                id="ac97a47f-f755-451e-956c-d1ed52403915",
                name="pos_gt",
                value_type="INTEGER",
                stage="GROUND_TRUTH",
                position=1,
                categorical=False,
                min_range=0,
                max_range=1,
                monitor_for_bias=False,
                is_unique=True,
                is_positive_predicted_attribute=False,
                attribute_link="pos_predicted_probability"
            ),
            ArthurAttribute(
                id="c2e626df-ee4c-403f-818b-f4f2d9180eab",
                name="neg_gt",
                value_type="INTEGER",
                stage="GROUND_TRUTH",
                position=2,
                categorical=False,
                min_range=0,
                max_range=1,
                monitor_for_bias=False,
                is_unique=True,
                is_positive_predicted_attribute=False,
                attribute_link="neg_predicted_probability"
            ),
            ArthurAttribute(
                id="d4f04345-c92c-4959-9d92-7fe35924efeb",
                name="neg_predicted_probability",
                value_type="FLOAT",
                stage="PREDICTED_VALUE",
                position=2,
                categorical=False,
                min_range=0,
                max_range=1,
                monitor_for_bias=False,
                is_unique=False,
                is_positive_predicted_attribute=False,
                attribute_link="neg_gt"
            ),
        ],
        "tags": [],
        "classifier_threshold": 0.5,
        "explainability": {
            "enabled": False
        },
        "is_batch": False,
        "text_delimiter": ","
    }
    yield ArthurModel(client=new_requests_client(url="test.ai", offline=True), **binary_classification_model)



from typing import Optional
from datetime import datetime

from arthurai.common.exceptions import arthur_excepted, UserValueError, MethodNotApplicableError

class BiasMetrics(object):
    def __init__(self, arthur_model):

        self.model = arthur_model 

    @arthur_excepted("Failed to get group-conditional confusion matrices")
    def group_confusion_matrices(self, attr_name: str, 
                           pred_property: Optional[str] = None,
                           batch_id: Optional[str] = None,
                           start_time: Optional[datetime] = None,
                           end_time: Optional[datetime] = None,
                           return_by_metric=True):
        """
        Get group-conditional confusion matrices for all inferences, with the option to filter for a `batch_id` or
        a particular chunk of time. Currently only supports calculating bias with respect to a single sensitive
        attribute at a time. This method handles both binary and multiclass models. 

        :param attr_name: The name of the sensitive attribute to get metrics for.
        :param pred_property: For multiclass models, the predicted label to get the confusion matrix for. 
        :param batch_id: Optional filter to limit calculations to a specific batch ID. 
        :param start_time: Optional filter to limit calculations to inferences after a specific timestamp.
        :param end_time: Optional filter to limit calculations to inferences before a specific timestamp.
        :param return_by_metric: Whether the returned dictionary of results should be keyed by metric or by sensitive attribute value.

        :return: Either a dict of the form `{ metric: { sens_1: val1, sens_2: val2 }, ...}` (default) or its reverse. 
        """

        # check whether model has any attributes marked as `monitor_for_bias`
        if not self.model.check_has_bias_attrs():
            raise MethodNotApplicableError("This model does not have any attributes set to monitor for bias.")

        if not self.model.check_attr_is_bias(attr_name):
            raise MethodNotApplicableError("This attribute is not set to monitor for bias.")

        if batch_id and not self.model.is_batch:
            raise UserValueError("This is not a batch model; remove the `batch_id` argument.")

        pos_attr = self.model.get_positive_predicted_class()

        if pos_attr:
            q = _bias_cm_query(attr_name, batch_id=batch_id, 
                                            classifier_threshold=self.model.classifier_threshold,
                                            start_time=start_time,
                                            end_time=end_time)
        elif not pred_property:
            raise UserValueError("For a multiclass model, the `pred_property` parameter must be specified.")
        else:
            q = _bias_multiclass_cm_query(attr_name, pred_property=pred_property,
                                            batch_id=batch_id,
                                            start_time=start_time,
                                            end_time=end_time)

        resp = self.model.query(q)

        sens_keyed = {x[attr_name]: x['confusion_matrix'] for x in resp}

        if return_by_metric:
            metr_keys = ['accuracy_rate', 'balanced_accuracy_rate', 'f1', 'false_negative_rate', 'false_positive_rate', 'precision', 'true_negative_rate', 'true_positive_rate']
            return {x: {s: sens_keyed[s][x]for s in sens_keyed.keys()} for x in metr_keys} # reverse the dictionary 

        return sens_keyed


    @arthur_excepted("Failed to get group-conditional positivity rates")
    def group_positivity_rates(self, attr_name: str,
                            batch_id: str = None,
                            start_time: datetime = None,
                            end_time:datetime = None):
        """
        Get group-conditional positivity rates for all inferences, with the option to filter for a `batch_id` or
        a particular chunk of time. Currently only supports calculating bias with respect to a single sensitive
        attribute at a time. This method currently only handles binary models. 

        :param attr_name: The name of the sensitive attribute to get metrics for.
        :param batch_id: Optional filter to limit calculations to a specific batch ID. 
        :param start_time: Optional filter to limit calculations to inferences after a specific timestamp.
        :param end_time: Optional filter to limit calculations to inferences before a specific timestamp.

        :return: A dict of the form {sens_1: pos_1, sens_2: pos_2, ...}
        """

        # check whether model has any attributes marked as `monitor_for_bias`
        if not self.model.check_has_bias_attrs():
            raise MethodNotApplicableError("This model does not have any attributes set to monitor for bias.")

        if not self.model.check_attr_is_bias(attr_name):
            raise MethodNotApplicableError("This attribute is not set to monitor for bias.")
        
        if batch_id and not self.model.is_batch:
            raise UserValueError("This is not a batch model; remove the `batch_id` argument.")

        pos_attr = self.model.get_positive_predicted_class()

        if pos_attr:
            q = _bias_pr_query(attr_name, pred_property=pos_attr,
                                            batch_id=batch_id, 
                                            classifier_threshold=self.model.classifier_threshold,
                                            start_time=start_time,
                                            end_time=end_time)
        else:
            raise UserValueError("Positivity rates are currently only supported for binary classifiers.")

        resp = self.model.query(q)
        return {x[attr_name]: x['positive_rate'] for x in resp}

    # ALIAS METHOD NAMES
    demographic_parity = group_positivity_rates


def _bias_cm_query(sens_attr, classifier_threshold=0.5,  batch_id=None, start_time=None, end_time=None):
    """
    Get group-conditional `metric` (either `confusionMatrixRate` or `rate`) for BINARY classifier
    If batch_id or start/end times are None, then do this over all batches / all time
    Query is identical for streaming
    """

    query = {
        "select": [
            {
                "property": sens_attr
            },
            {
                "function": "confusionMatrixRate",
                "alias": "confusion_matrix",
                "parameters": {
                "threshold": classifier_threshold
                }
            }
            ],
        "group_by": [
            {
                "property": sens_attr
            }
        ]
        }

    query = _add_filters(query, batch_id, start_time, end_time)

    return query

def _bias_multiclass_cm_query(sens_attr, pred_property, batch_id=None, start_time=None, end_time=None):
    """
    Get group-conditional confusion matrix for MULTICLASS classifier (`rate` doesn't apply here, since there is no single threshold)
    If batch_id or start/end times are None, then do this over all batches / all time
    Query is identical for streaming
    """

    query =   {
        "select": [
        {
            "property": sens_attr
        },
        {
            "function": "confusionMatrixRateMulticlass",
            "alias": "confusion_matrix",
            "parameters": {
                "predicted_property": pred_property
            }
        }
        ],
        "group_by": [
        {
            "property": sens_attr
        }
        ]
    }

    query = _add_filters(query, batch_id, start_time, end_time)

    return query

def _bias_pr_query(sens_attr, pred_property, classifier_threshold=0.5, batch_id=None, start_time=None, end_time=None):
    """
    Get group-conditional positivity rates for BINARY classifier
    If batch_id or start/end times are None, then do this over all batches/all time
    Query is identical for streaming
    """

    query = {
        "select": [
            {
                "property": sens_attr
            },
            {
                "function": "rate",
                "alias": "positive_rate",
                "parameters": {
                    "property": pred_property,
                    "comparator": "gte",
                    "value": classifier_threshold
                }
            }
        ], 
        "group_by": [{
            "property": sens_attr
        }]
    }

    query = _add_filters(query, batch_id, start_time, end_time)
    
    return query

def _add_filters(query, batch_id = None, start_time=None, end_time=None):
    if batch_id:
        query["filter"] = [
            {
                "property": "batch_id",
                "comparator": "eq",
                "value": batch_id
            }           
        ]
    
    if start_time:
        query["filter"] = [
            {
                "property": "inference_timestamp",
                "comparator": "gte",
                "value": start_time.strftime("%Y-%m-%dT%H:%M:%SZ")
            }
        ]

    if end_time:
        query["filter"] = [
            {
                "property": "inference_timestamp",
                "comparator": "lte",
                "value": end_time.strftime("%Y-%m-%dT%H:%M:%SZ")
            }
        ]

    return query



from collections import OrderedDict

import pandas as pd
import pytest

from arthurai import ArthurAttribute, ArthurModel
from arthurai.common.constants import ValueType, Stage, InputType, OutputType
from arthurai.core.attributes import AttributeCategory

from tests.fixtures.models.regression import INT_INPUT_ATTR, FLOAT_NONINPUT_ATTR, INT_INPUT, FLOAT_NONINPUT, \
    TABULAR_MODEL_ID, PRED as REG_PRED, GROUND_TRUTH as REG_GROUND_TRUTH
from tests.fixtures.mocks import client

DOG_GROUND_TRUTH = "dog_gt"
DOG_PRED = "dog_pred"
CAT_GROUND_TRUTH = "cat_gt"
CAT_PRED = "cat_pred"
IMPLICIT_NEG_PRED = "predicted_negative"
GROUND_TRUTH = "single_col_gt"

DOG_GROUND_TRUTH_ATTR = ArthurAttribute(name=DOG_GROUND_TRUTH, value_type=ValueType.Integer, stage=Stage.GroundTruth,
                                        position=0, attribute_link=DOG_PRED, categorical=True,
                                        categories=[AttributeCategory(value="0"), AttributeCategory(value="1")])
GENERIC_GROUND_TRUTH_ATTR = ArthurAttribute(name=REG_GROUND_TRUTH, value_type=ValueType.Integer, stage=Stage.GroundTruthClass,
                                            position=0, categorical=True, categories=[AttributeCategory(value="0"), AttributeCategory(value="1")])
GENERIC_PRED_ATTR = ArthurAttribute(name=REG_PRED, value_type=ValueType.Float, stage=Stage.PredictedValue, position=0,
                                    is_positive_predicted_attribute=True)
DOG_PRED_ATTR = ArthurAttribute(name=DOG_PRED, value_type=ValueType.Float, stage=Stage.PredictedValue, position=0,
                                attribute_link=DOG_GROUND_TRUTH, is_positive_predicted_attribute=True, min_range=0.,
                                max_range=1.)
CAT_GROUND_TRUTH_ATTR = ArthurAttribute(name=CAT_GROUND_TRUTH, value_type=ValueType.Integer, stage=Stage.GroundTruth,
                                        position=1, attribute_link=CAT_PRED, categorical=True,
                                        categories=[AttributeCategory(value="0"), AttributeCategory(value="1")])
CAT_GROUND_TRUTH_ATTR_UNLINKED = ArthurAttribute(
    name=CAT_GROUND_TRUTH, value_type=ValueType.Integer, stage=Stage.GroundTruth, position=1,
    categorical=True, categories=[AttributeCategory(value="0"), AttributeCategory(value="1")])
CAT_PRED_ATTR = ArthurAttribute(name=CAT_PRED, value_type=ValueType.Float, stage=Stage.PredictedValue, position=1,
                                attribute_link=CAT_GROUND_TRUTH, min_range=0., max_range=1.)

IMPLICIT_NEG_PRED_ATTR = ArthurAttribute(name=IMPLICIT_NEG_PRED, value_type=ValueType.Float, stage=Stage.PredictedValue,
                                         position=1, attribute_link=CAT_GROUND_TRUTH)
CAT_GROUND_TRUTH_ATTR_IMPLICIT_LINK = ArthurAttribute(
    name=CAT_GROUND_TRUTH, value_type=ValueType.Integer, stage=Stage.GroundTruth, position=1,
    categorical=True, categories=[AttributeCategory(value="0"), AttributeCategory(value="1")],
    attribute_link=IMPLICIT_NEG_PRED)
SINGLE_COL_GT_ATTR = ArthurAttribute(name=GROUND_TRUTH, value_type=ValueType.String, stage=Stage.GroundTruthClass,
                                     position=0, categorical=True,
                                     categories=[AttributeCategory(value="0"), AttributeCategory(value="1")])
UNLINKED_DOG_PRED = ArthurAttribute(name=DOG_PRED, value_type=ValueType.Float, stage=Stage.PredictedValue, position=0,
                                    is_positive_predicted_attribute=True, min_range=0., max_range=1.,
                                    gt_class_link="1")
UNLINKED_CAT_PRED = ArthurAttribute(name=CAT_PRED, value_type=ValueType.Float, stage=Stage.PredictedValue, position=1,
                                    min_range=0., max_range=1., gt_class_link="0")

PRED_TO_GT_MAP = OrderedDict()
PRED_TO_GT_MAP[DOG_PRED] = DOG_GROUND_TRUTH
PRED_TO_GT_MAP[CAT_PRED] = CAT_GROUND_TRUTH

MODEL_ATTRIBUTES_NO_NEG_PRED = [INT_INPUT_ATTR, FLOAT_NONINPUT_ATTR, DOG_PRED_ATTR,
                                DOG_GROUND_TRUTH_ATTR, CAT_GROUND_TRUTH_ATTR_UNLINKED]
MODEL_ATTRIBUTES = [INT_INPUT_ATTR, FLOAT_NONINPUT_ATTR, DOG_PRED_ATTR, CAT_PRED_ATTR,
                    DOG_GROUND_TRUTH_ATTR, CAT_GROUND_TRUTH_ATTR]
MODEL_ATTRIBUTES_IMPLICIT_NEG = [INT_INPUT_ATTR, FLOAT_NONINPUT_ATTR, DOG_PRED_ATTR, IMPLICIT_NEG_PRED_ATTR,
                                 DOG_GROUND_TRUTH_ATTR, CAT_GROUND_TRUTH_ATTR_IMPLICIT_LINK]
MODEL_ATTRIBUTES_GENERIC = [INT_INPUT_ATTR, FLOAT_NONINPUT_ATTR, GENERIC_PRED_ATTR,
                            GENERIC_GROUND_TRUTH_ATTR]
SINGLE_COL_MODEL_ATTRIBUTES = [INT_INPUT_ATTR, FLOAT_NONINPUT_ATTR, UNLINKED_DOG_PRED, UNLINKED_CAT_PRED,
                               SINGLE_COL_GT_ATTR]

SAMPLE_DATAFRAME = pd.DataFrame({
    INT_INPUT: [1, 2, 1, 2, 1, 2],
    FLOAT_NONINPUT: [1., 0.5, 1., .25, .6, .9],
    DOG_PRED: [0.6, 0.8, 0.2, 0.45, 0.94, 0.12],
    CAT_PRED: [0.4, 0.2, 0.8, 0.65, 0.06, 0.88],
    DOG_GROUND_TRUTH: [0, 1, 0, 1, 1, 0],
    CAT_GROUND_TRUTH: [1, 0, 1, 0, 0, 1]
})

SAMPLE_DATAFRAME_NO_NEG_PRED = SAMPLE_DATAFRAME.drop(columns=[CAT_PRED])

SINGLE_COLUMN_DATAFRAME = pd.DataFrame({
    INT_INPUT: [1, 2, 1, 2, 1, 2],
    FLOAT_NONINPUT: [1., 0.5, 1., .25, .6, .9],
    DOG_PRED: [0.6, 0.8, 0.2, 0.45, 0.94, 0.12],
    CAT_PRED: [0.4, 0.2, 0.8, 0.65, 0.06, 0.88],
    GROUND_TRUTH: ["0", "1", "0", "1", "1", "0"],
})

UNINITIALIZED_MODEL_DATA = {
        "partner_model_id": "",
        "input_type": InputType.Tabular,
        "output_type": OutputType.Multiclass,
        "display_name": "",
        "description": ""
}

MODEL_ID = "c4ea58b6-4ec7-43b2-94d3-786eccb2a492"

MODEL_DATA_WITH_ATTRIBUTES = {
    **UNINITIALIZED_MODEL_DATA,
    "attributes": MODEL_ATTRIBUTES
}

GENERIC_MODEL_DATA = {
    **UNINITIALIZED_MODEL_DATA,
    "attributes": MODEL_ATTRIBUTES_GENERIC
}

MODEL_DATA_WITH_ATTRIBUTES_IMPLICIT_NEG = {
    "id": MODEL_ID,
    "partner_model_id": "mclass123",
    "input_type": InputType.Tabular,
    "output_type": OutputType.Multiclass,
    "attributes": MODEL_ATTRIBUTES_IMPLICIT_NEG
}


@pytest.fixture
def initial_biclass_model(client):
    model = ArthurModel(client=client.client, **UNINITIALIZED_MODEL_DATA)
    return model


@pytest.fixture
def biclass_model(client):
    model = ArthurModel(client=client.client, **GENERIC_MODEL_DATA)
    model.id = MODEL_ID
    return model



import pytest
import pathlib
import os
import math
import pandas as pd
import tempfile
import shutil
from typing import Iterable

from arthurai.core.data_service import DatasetService
from arthurai.core import util


IMAGE_COL_NAME = 'image'
EXTRA_COL_NAME = 'extra'  # dummy column, ensure we aren't loosing columns

IMAGE_PATH = os.path.join(pathlib.Path(__file__).parent.absolute(), "data", "test_image.png")
IMAGE_SIZE = os.path.getsize(IMAGE_PATH)

# (chunk size, file sizes)
chunk_image_set_cases = [pytest.param(3_000, [4_000], id="single-file"),
                         pytest.param(3_000, [400 for _ in range(10)], id="ten-files"),
                         pytest.param(3_000, [3_600, 400], id="diff-sized-files")]


@pytest.mark.parametrize("chunk_size, file_sizes", chunk_image_set_cases)
def test_chunk_parquet_image_set(chunk_size: int, file_sizes: Iterable[int]):
    image_counts_by_file = [math.ceil(target_size / IMAGE_SIZE) for target_size in file_sizes]
    # create temporary directory to generate input files in, with auto-cleanup
    with tempfile.TemporaryDirectory() as source_directory:
        # create the input files
        for i in range(len(image_counts_by_file)):
            # generate the image column data
            num_images = image_counts_by_file[i]
            image_col_data = [IMAGE_PATH for _ in range(num_images)]
            df = pd.DataFrame({IMAGE_COL_NAME: image_col_data, EXTRA_COL_NAME: 1})
            # and write it to parquet
            df.to_parquet(os.path.join(source_directory, f'test-{i}.parquet'))

        # chunk the input files
        output_directory = DatasetService.chunk_parquet_image_set(source_directory, IMAGE_COL_NAME,
                                                                  max_image_data_bytes=chunk_size)

        # validate no data loss in chunked files
        chunked_files = util.retrieve_parquet_files(output_directory)
        total_chunked_images = 0
        for chunked_file in chunked_files:
            df = pd.read_parquet(chunked_file)
            total_chunked_images += len(df)

            # don't lose columns
            assert EXTRA_COL_NAME in df

            # image chunk can be at most image_size + ~chunk_size, we check size after adding file
            cur_chunk_size = len(df) * IMAGE_SIZE
            assert cur_chunk_size - chunk_size <= IMAGE_SIZE
        assert total_chunked_images == sum(image_counts_by_file)

        # cleanup output directory created by the chunking function
        shutil.rmtree(output_directory)




"""SDK Version 2 backward compatibility

This convenience module provides compatibility with the ArthurAI SDK version 2.x, but is considered deprecated. Please
update your import paths according to the following:

* arthurai.client.apiv3.InputType → arthurai.common.constants.InputType
* arthurai.client.apiv3.OutputType → arthurai.common.constants.OutputType
* arthurai.client.apiv3.ValueType → arthurai.common.constants.ValueType
* arthurai.client.apiv3.Stage → arthurai.common.constants.Stage
* arthurai.client.apiv3.TextDelimiter → arthurai.common.constants.TextDelimiter
* arthurai.client.apiv3.ListableStrEnum → arthurai.common.constants.ListableStrEnum

* arthurai.client.apiv3.AttributeCategory → arthurai.core.attributes.AttributeCategory
* arthurai.client.apiv3.AttributeBin → arthurai.core.attributes.AttributeBin

* arthurai.client.apiv3.arthur_explainer → arthurai.explainability.arthur_explainer
* arthurai.client.apiv3.decorators → arthurai.core.decorators
* arthurai.client.apiv3.explanation_packager → arthurai.explainability.explanation_packager

"""

import logging

from arthurai.common.constants import InputType, OutputType, ValueType, Stage, TextDelimiter, ListableStrEnum
from arthurai.core import decorators
from arthurai.core.attributes import AttributeCategory, AttributeBin
from arthurai.explainability import arthur_explainer, explanation_packager


logger = logging.getLogger(__name__)
logger.warning("DEPRECATION WARNING: The arthurai.client.apiv3 module is provided for backward compatibility, but is "
               "deprecated and will be removed. Please update your import paths, for details see the "
               "arthurai.client.apiv3 module documentation")



from arthurai.explainability.explanation_packager import ExplanationPackager
from arthurai.core.models import ArthurModel
from arthurai.common.constants import InputType, OutputType

import pkg_resources
import os

#TODO Katie figure out if tests are needed here for image input type (if image loading logic is added etc)
def test_validate_requirements() -> None:
    # build dict of installed packages and versions
    
    # gather user env, skipping sklearn, since we don't want to grab that when we build test req file
    cur_env = [
        (pkg.project_name, pkg.version)
        for pkg in pkg_resources.working_set
        if pkg.project_name != 'scikit-learn'
    ]

    # generate a requirements file that should succeed
    req_path = '/tmp/user_requirements.txt'
    with open(req_path, 'w') as f:
        test_package, test_version = cur_env[0]
        f.write(f"{test_package}=={test_version}")
    
    model = ArthurModel("123", InputType.Tabular, OutputType.Regression)
    packager = ExplanationPackager(model, requirements_file=req_path)

    # validate no errors
    errors = packager.validate_requirements_add_sklearn()
    assert len(errors) == 0

    # validate sklearn was added
    # we should always have sklearn in testing environment, since bunlded with shap
    with open(req_path, 'r') as f:
        req_lines = f.readlines()

    assert len(req_lines) == 2
    assert 'scikit-learn' in req_lines[1]

    # change version in req file to fail
    with open(req_path, 'w+') as f:
        test_package, test_version = cur_env[0]
        test_version = "-1.-1.-1"
        f.write(f"{test_package}=={test_version}")

    packager = ExplanationPackager(model, requirements_file=req_path)

    # validate no errors
    errors = packager.validate_requirements_add_sklearn()
    assert len(errors) == 1


def test_ignore_hidden_files():
    """Tests to make sure we don't bundle up hidden files in a given directory"""
    # Create a requirements file
    cur_env = [
        (pkg.project_name, pkg.version)
        for pkg in pkg_resources.working_set
        if pkg.project_name != 'scikit-learn'
    ]

    req_path = '/tmp/user_requirements.txt'
    with open(req_path, 'w') as f:
        test_package, test_version = cur_env[0]
        f.write(f"{test_package}=={test_version}")

    # Create a hidden file directory
    project_directory_path = '/tmp/.venv'
    if not os.path.exists(project_directory_path):
        os.makedirs(project_directory_path)

    model = ArthurModel("123", InputType.Tabular, OutputType.Regression)
    packager = ExplanationPackager(model, requirements_file=req_path, project_directory='/tmp')

    # Ensure the hidden directory is in ignore dirs (which will be handled by ignore dirs logic)
    assert('.venv' in packager.ignore_dirs)


from datetime import datetime
import logging
from typing import List, Dict, Any, Union, Iterable, Collection
from collections.abc import Iterable as IterableABC

import pandas as pd
import pytz
import shortuuid

from arthurai import ArthurAttribute
from arthurai.common.constants import Stage
from arthurai.common.exceptions import InternalValueError, UserValueError, UserTypeError
from arthurai.core.util import intersection_is_non_empty


INFERENCE_STAGES = {Stage.ModelPipelineInput, Stage.PredictFunctionInput, Stage.NonInputData, Stage.PredictedValue}
RESERVED_COLUMNS = {"inference_timestamp", "partner_inference_id", "batch_id", "ground_truth_timestamp",
                    "inference_data", "ground_truth_data"}
FALSE_DEFAULT_IGNORE_JOIN_ERRORS = False

logger = logging.getLogger(__name__)


def parse_stage_attributes(data: Union[List[Dict[str, Any]], Dict[str, List[Any]],
                                       pd.DataFrame, Iterable[Any]],
                           attributes: List[ArthurAttribute],
                           stage: Stage):
    """Parses data for a single stage into the standard List of Dicts format.

    If the stage contains only a single attribute, data can be a single list-like column.

    .. seealso::
        Similar to dataframe_like_to_list_of_dicts, but an expected column (attribute)-aware and supports single columns

    :param data:
    :param attributes:
    :param stage:
    :return: parsed data in List of Dicts formt
    """
    expected_attributes = {attr.name for attr in attributes if attr.stage == stage}

    # parsing
    if isinstance(data, list) and isinstance(data[0], dict):
        pass
    elif isinstance(data, dict):
        data = pd.DataFrame(data).to_dict(orient='records')
    elif isinstance(data, pd.DataFrame):
        data = data.to_dict(orient='records')
    elif isinstance(data, IterableABC):
        if len(expected_attributes) != 1:
            raise UserValueError(f"Only a single column was provided for {stage} data but there are "
                                 f"{len(expected_attributes)} expected attributes: {expected_attributes}")
        attr = next(iter(expected_attributes))
        data = [{attr: val} for val in data]
    else:
        raise UserTypeError(f"Cannot parse {stage} data of type {type(data)}, should be list of dicts, dict of "
                            f"lists, DataFrame, or Sequence of values")

    return data


def add_predictions_or_ground_truth(inference_data: List[Dict[str, Any]],
                                    new_data: Union[List[Dict[str, Any]], Dict[str, List[Any]],
                                                    pd.DataFrame, Iterable[Any]],
                                    attributes: List[ArthurAttribute],
                                    stage: Stage):
    """
    Add prediction or ground truth data to inference data *in place*.
    :param inference_data: the inference data as a List of Dicts as expected by the Arthur API
    :param new_data: the new data to add in, as a List of Dicts, Dict of Lists, DataFrame, or Sequence
    (if a single column)
    :param attributes: the model's attributes
    :param stage: the Stage of the new data, either PredictedValue or GroundTruth
    :return: None (modifies inference_data in place)
    """
    # initial validation and config
    if stage == Stage.GroundTruth or stage == Stage.GroundTruthClass:
        target_field = "ground_truth_data"
    elif stage == Stage.PredictedValue:
        target_field = "inference_data"
    else:
        raise InternalValueError(f"Unexpected stage {stage}, should be GroundTruth or PredictedValue")

    # dataframe being interpreted as Any but adding pandas stubs break a lot of typing so ignoring for now
    if len(new_data) == 0:  # type: ignore
        if len(inference_data) == 0:
            return inference_data
        else:
            raise UserValueError(f"Cannot add empty {stage} data to non-empty inference data")

    # parse new data into common format
    new_data = parse_stage_attributes(new_data, attributes, stage)

    # add new data
    if len(new_data) != len(inference_data):
        raise UserValueError(f"Size of {stage} data ({len(new_data)} does not match size of inference data "
                             f"{len(inference_data)}")
    for i in range(len(inference_data)):
        if target_field not in inference_data[i].keys():
            inference_data[i][target_field] = {}
        for new_key in new_data[i].keys():
            inference_data[i][target_field][new_key] = new_data[i][new_key]


def nest_inference_and_ground_truth_data(data: List[Dict[str, Any]], attributes: List[ArthurAttribute]) -> \
        List[Dict[str, Any]]:
    """
    Reformat List of Dicts inference data to nest inference and ground truth data as expected by the Arthur API.
    For example:

    .. code-block:: python

        [
            {
                "input_attr_1": 1.0,
                "prediction_1": 0.95,
                "inference_timestamp": "2021-06-03T19:44:33.169334+00:00",
                "ground_truth_1": 1,
                "ground_truth_timestamp": "2021-06-03T19:44:56.892019+00:00"
            }
        ]

    Will become:

    .. code-block:: python

        [
            {
                "inference_data":
                {
                    "input_attr_1": 1.0,
                    "prediction_1": 0.95
                },
                "ground_truth_data":
                {
                    "ground_truth_1": 1
                },
                "inference_timestamp": "2021-06-03T19:44:33.169334+00:00",
                "ground_truth_timestamp": "2021-06-03T19:44:56.892019+00:00"
            }
        ]

    :param data: the input data to reformat, either already nested or flat
    :param attributes: the model's attributes
    :return: the nested data
    """
    if len(data) == 0:
        return []

    inference_attrs = {attr.name for attr in attributes if attr.stage in INFERENCE_STAGES}
    gt_attrs = {attr.name for attr in attributes
                if attr.stage == Stage.GroundTruth or attr.stage == Stage.GroundTruthClass}

    # ensure the data is not mixed nested and un-nested
    for i in range(len(data)):
        row = data[i]
        if "inference_data" in row.keys() and intersection_is_non_empty(inference_attrs, row.keys()):
            raise UserValueError(f"Inference data should be nested or non-nested, not mixed. Row {i} contains "
                                 "'inference_data' field and inference attribute fields.")
        if "ground_truth_data" in row.keys() and intersection_is_non_empty(gt_attrs, row.keys()):
            raise UserValueError(f"Ground truth data should be nested or non-nested, not mixed. Row {i} contains "
                                 "'ground_truth_data' field and ground truth attribute fields.")

    # generate nested data
    nested_data = []
    for i in range(len(data)):
        row = data[i]
        nested_row: Dict[str, Any] = {}
        # if the data needs to be nested create the wrapper dicts, otherwise validate the type
        if 'inference_data' not in row.keys():
            if intersection_is_non_empty(inference_attrs, row.keys()):
                nested_row['inference_data'] = {}
        elif not isinstance(row['inference_data'], dict):
            raise UserTypeError(f"'inference_data' field present but is of type {type(row['inference_data'])}, "
                                f"should be dict")
        if 'ground_truth_data' not in row.keys():
            if intersection_is_non_empty(gt_attrs, row.keys()):
                nested_row['ground_truth_data'] = {}
        elif not isinstance(row['ground_truth_data'], dict):
            raise UserTypeError(f"'ground_truth_data' field present but is of type "
                                f"{type(row['ground_truth_data'])}, should be dict")

        # copy the data into the new struct
        for key in row.keys():
            if key in inference_attrs:
                nested_row['inference_data'][key] = row[key]
            elif key in RESERVED_COLUMNS:
                # copy reserved dict columns since they may be nested
                if isinstance(row[key], dict):
                    nested_row[key] = row[key].copy()
                else:
                    nested_row[key] = row[key]
            elif key in gt_attrs:
                nested_row['ground_truth_data'][key] = row[key]
            else:
                raise UserValueError(f"Unexpected key {key} on row {i} of data")
        nested_data.append(nested_row)

    return nested_data


def add_inference_metadata_to_dataframe(df: pd.DataFrame,
                                        model_attributes: Collection[ArthurAttribute],
                                        ignore_join_errors: bool = FALSE_DEFAULT_IGNORE_JOIN_ERRORS) -> pd.DataFrame:
    """Adds timestamp and/or partner_inference_id fields to the DataFrame.

    :param df: DataFrame to add metadata to
    :param model_attributes: attributes of the model
    :param ignore_join_errors: if True, allow inference data without ``partner_inference_id`` or ground truth data
    :return: the input DataFrame if no updates are needed, otherwise a shallow copy with the new columns
    :raises UserValueError: if inference data is supplied without ``partner_inference_id`` or ground truth data, and
        ``ignore_join_errors`` is False.
    """
    # determine whether the dataset contains inference data, ground truth data, or both
    attributes_by_name = {attr.name: attr for attr in model_attributes}
    contains_inference_columns = False
    contains_ground_truth_columns = False
    for col in df.columns:
        if col in attributes_by_name.keys():
            attr = attributes_by_name[col]
            if attr.stage in INFERENCE_STAGES:
                contains_inference_columns = True
            elif attr.stage == Stage.GroundTruth or attr.stage == Stage.GroundTruthClass:
                contains_ground_truth_columns = True
        if contains_inference_columns and contains_ground_truth_columns:
            break
    if (not contains_inference_columns) and (not contains_ground_truth_columns):
        raise UserValueError(f"No inference or ground truth data was found in the DataFrame with columns "
                             f"{list(df.columns)}")

    # determine what fields need to be added so that we can avoid copying the DataFrame if there's no work to do
    add_inference_timestamp, add_gt_timestamp, add_partner_ids = False, False, False
    if contains_inference_columns and 'inference_timestamp' not in df.columns:
        add_inference_timestamp = True
    if contains_ground_truth_columns and 'ground_truth_timestamp' not in df.columns:
        add_gt_timestamp = True
    if 'partner_inference_id' not in df.columns:
        # ensure that the customer isn't only supplying ground truth
        if contains_ground_truth_columns and (not contains_inference_columns):
            raise UserValueError(f"Cannot send ground truth data without `partner_inference_id` field to join to "
                                 "existing inference data. Please supply the `partner_inference_id` of the already-"
                                 "uploaded inferences or supply inference data with your ground truth data.")
        add_partner_ids = True

    if not (add_inference_timestamp or add_gt_timestamp or add_partner_ids):
        return df

    if (contains_inference_columns and (not contains_ground_truth_columns) and add_partner_ids
            and (not ignore_join_errors)):
        raise UserValueError(
            "You are sending inference data without `partner_inference_id`s or ground truth data, it will be difficult "
            "to append ground truth data at a later time. If you wish to append ground truth data at a later time "
            "please supply a `partner_inference_id` field to your inferences or use the synchronous send_inferences() "
            "method to auto-generate them, and retain the generated values. If you do not intend to send ground truth "
            "data you may override this error with the 'ignore_join_errors=True' flag.")

    # copy the dataframe and add the needed fields
    df = df.copy()
    current_timestamp = datetime.now(pytz.utc)
    if add_inference_timestamp:
        df['inference_timestamp'] = current_timestamp
        logger.info("inference_timestamp field was not supplied in the DataFrame, so the current time was populated")
    if add_gt_timestamp:
        df['ground_truth_timestamp'] = current_timestamp
        logger.info("ground_truth_timestamp field was not supplied in the DataFrame, so the current time was populated")
    if add_partner_ids:
        df['partner_inference_id'] = [str(shortuuid.uuid()) for _ in range(len(df))]

    return df



import json
from datetime import datetime
from http import HTTPStatus
from typing import Dict, Any, Union, List, Tuple
from unittest.mock import MagicMock

import numpy as np
import pandas as pd
import pytz
import responses

from arthurai import ArthurAttribute
from arthurai.core.base import ArthurBaseJsonDataclass
from tests.fixtures.mocks import BASE_URL

from unittest import TestCase

CONTENT_TYPE_JSON = 'application/json'
OK_DEFAULT_STATUS = HTTPStatus.OK


def mock_response(request_type: str, append_url: str, response_body: Union[Dict, List, str, bytes],
                  status=OK_DEFAULT_STATUS, headers: Dict[str, str] = None):
    """
    Adds a mock response

    :param request_type: the type of HTTP request expected (e.g. GET)
    :param append_url: the url to the base url
    :param response_body: the response to return from the mock API request
    :param status: status code of the response
    :param headers: content type header of the returned response

    :return: None
    """
    if headers is None:
        headers = {}
    # responses likes the content type header to be specified as a param
    content_type = headers.pop('Content-Type', CONTENT_TYPE_JSON)

    parsed_body: Union[str, bytes] = ""
    if isinstance(response_body, list) or isinstance(response_body, dict):
        parsed_body = json.dumps(response_body, default=ArthurBaseJsonDataclass.to_dict)
    else:
        parsed_body = response_body

    responses.add(request_type, BASE_URL + append_url, parsed_body, status=status, headers=headers,
                  content_type=content_type)


def mock_post(append_url: str, response_body: Union[Dict, List, str, bytes], status=OK_DEFAULT_STATUS,
              headers: Dict[str, str] = None):
    """
    Adds a mock 200 response to a POST request

    :param append_url: the url to the base url
    :param response_body: the response to return from the mock API request
    :param status: status code of the response
    :param headers: content type header of the returned response

    :return: None
    """
    mock_response(responses.POST, append_url, response_body, status=status, headers=headers)


def mock_patch(append_url: str, response_body: Union[Dict, List, str, bytes], status=OK_DEFAULT_STATUS,
               headers: Dict[str, str] = None):
    """
    Adds a mock 200 response to a PATCH request

    :param append_url: the url to the base url
    :param response_body: the response to return from the mock API request
    :param status: status code of the response
    :param headers: content type header of the returned response

    :return: None
    """
    mock_response(responses.PATCH, append_url, response_body, status=status, headers=headers)


def mock_get(append_url: str, response_body: Union[Dict, List, str, bytes], status=200, headers: Dict[str, str] = None):
    """
    Adds a mock 200 response

    :param append_url: the url to the base url
    :param response_body: the response to return from the mock API request
    :param status: status code of the response
    :param headers: headers of the returned response

    :return: None
    """
    mock_response(responses.GET, append_url, response_body, status=status, headers=headers)


class MockShortUUID:
    def __init__(self, ids=None):
        if ids is None:
            self.ids = []
        else:
            self.ids = ids
        self.next_idx = 0

    def call_count(self):
        return self.next_idx

    def next(self):
        value = self.ids[self.next_idx]
        self.next_idx += 1
        return value


class MockDatetime:
    def __init__(self, return_vals=None, return_raw_strings=False):
        if return_vals is None or len(return_vals) == 0:
            self.timestamps = []
        else:
            if isinstance(return_vals[0], datetime):
                self.timestamps = return_vals
            elif isinstance(return_vals[0], str):
                mocks = []
                for ts in return_vals:
                    if return_raw_strings:
                        mocks.append(ts)
                    else:
                        dt = MagicMock(datetime)
                        dt.isoformat.return_value = ts
                        mocks.append(dt)
                self.timestamps = mocks
        self.next_idx = 0

    def call_count(self):
        return self.next_idx

    def next(self, tz):
        if tz != pytz.utc:
            raise ValueError("unexpected timezone")
        value = self.timestamps[self.next_idx]
        self.next_idx += 1
        return value


def assert_kwargs_equal(actual: Dict[str, Any], expected: Dict[str, Any]):
    assert actual.keys() == expected.keys()
    for key in expected.keys():
        if isinstance(actual[key], (pd.Series, pd.DataFrame)):
            assert actual[key].equals(expected[key])
        elif isinstance(actual[key], np.ndarray):
            assert np.equal(actual[key], expected[key]).all()
        else:
            assert actual[key] == expected[key]


def assert_attributes_equal(expected: List[ArthurAttribute], actual: List[ArthurAttribute]):
    # create a map from attribute name to the index in the list
    expected_name_to_index = {attr.name: i for i, attr in enumerate(expected)}
    actual_name_to_index = {attr.name: i for i, attr in enumerate(actual)}
    if len(expected_name_to_index) != len(expected) or len(actual_name_to_index) != len(actual):
        raise ValueError("attributes in lists have duplicate names")

    # create pairs of the matching attributes to compare to each other
    expected_actual_pairs: List[Tuple[ArthurAttribute, ArthurAttribute]] = []
    missing_attrs = []
    for expected_attr in expected:
        if expected_attr.name not in actual_name_to_index.keys():
            missing_attrs.append(expected_attr.name)
        else:
            expected_actual_pairs.append((expected_attr, actual[actual_name_to_index[expected_attr.name]]))

    extra_attrs = []
    for actual_attr in actual:
        if actual_attr.name not in expected_name_to_index.keys():
            extra_attrs.append(actual_attr.name)
    assert len(missing_attrs) == 0, f"actual attributes list is missing: {', '.join(missing_attrs)}"

    # iterate through pairs, comparing their values
    failures = []
    for expected_attr, actual_attr in expected_actual_pairs:
        # assert each field is the same, but categories field can have different order
        try:
            assert expected_attr.name == actual_attr.name
            assert expected_attr.stage == actual_attr.stage
            assert expected_attr.id == actual_attr.id
            assert expected_attr.position == actual_attr.position
            assert expected_attr.categorical == actual_attr.categorical
            assert expected_attr.min_range == actual_attr.min_range
            assert expected_attr.max_range == actual_attr.max_range
            assert expected_attr.monitor_for_bias == actual_attr.monitor_for_bias
            if expected_attr.categories is None:
                assert actual_attr.categories is None
            else:
                TestCase().assertCountEqual(expected_attr.categories, actual_attr.categories)
            assert expected_attr.bins == actual_attr.bins
            assert expected_attr.is_unique == actual_attr.is_unique
            assert expected_attr.attribute_link == actual_attr.attribute_link
        except AssertionError as e:
            failures.append(f"Attribute {expected_attr.name} does not match expected value.\nActual:   {actual_attr}\n"
                            f"Expected: {expected_attr}\nError: {e}")
    assert len(failures) == 0, "\n".join(failures)



import pytest

from tests.test_request_models.fixtures import AlertRuleDict_1, \
    AlertRuleJson_1, AlertRuleObject_1, AlertRuleObject_2, AlertRuleDict_2, AlertRuleJson_2, \
    AlertRuleObject_3, AlertRuleDict_3, AlertRuleJson_3
from arthurai.core.alerts import AlertRule


class TestAlertRule:

    @pytest.mark.parametrize("alert_rule_object, alert_rule_dict, alert_rule_json", [
        (AlertRuleObject_1, AlertRuleDict_1, AlertRuleJson_1),
        (AlertRuleObject_2, AlertRuleDict_2, AlertRuleJson_2),
        (AlertRuleObject_3, AlertRuleDict_3, AlertRuleJson_3)
    ])
    def test_alert_rule_model(self, alert_rule_object, alert_rule_dict, alert_rule_json):
        assert alert_rule_object.to_json() == alert_rule_json
        assert AlertRule.from_json(alert_rule_json) == alert_rule_object

        assert alert_rule_object.to_dict() == alert_rule_dict
        assert AlertRule.from_dict(alert_rule_dict) == alert_rule_object
